import sys
import string
import math
import os
from Bio import SeqIO


class Shortread:
	def __init__(self, hairpin_id, offset, sequence, count, mature_id, position, prime, mat_offset, highest_prime):
		self.hairpin_id = hairpin_id
		self.offset = offset
		self.sequence = sequence
		self.count = count
		self.mature_id = mature_id
		self.position = position
		self.prime = prime
		self.mat_offset = mat_offset
		self.highest_prime = highest_prime

class MicroRNA:
	def __init__(self, hairpin_id, miRNA_id, count, offset, sequence, prime):
		self.hairpin_id = hairpin_id
		self.miRNA_id = miRNA_id
		self.count = count
		self.offset = offset
		self.sequence = sequence
		self.prime = prime
		self.expressed = True
		self.highest_prime = None
		self.mature = False
		self.hasShortreads = False

	def setUnexpressed(self):
		self.expressed = False

	def setHighestPrime(self, highest_prime):
		self.highest_prime = True if self.prime==highest_prime else False

	def setMature(self, matureID):
		if self.miRNA_id==matureID:
			self.mature = True 

	def setShortread(self):
		self.hasShortreads = True

class Hairpin:
	def __init__(self, hairpin_id, sequence, foldstructure):
		self.hairpin_id = hairpin_id
		self.sequence = sequence
		self.foldstructure = foldstructure
		self.expressed = False

	def setExpressed(self):
		self.expressed = True


def main(refFile, dotBracketFile, matFile, sampFile, name, total_name):
	"""
	Main method, calls all functions necessary.
	"""
	global hairpin_map, shortreads, shortread_map, microRNA_map, bucketSize, unannotated
	# Initialize global data structures
	hairpin_map = dict()
	shortreads = dict()
	shortread_map = dict()
	microRNA_map = dict()
	unannotated = dict()
	bucketSize = 5

	#Load input data
	readHairpins(refFile, dotBracketFile) 
    # dotbracket = secondary structure rna in dotbracket format
    # matfile = mature miRNA in fasta format <mature.fa>
    # sampFile = sample file ..
    # refFile? =
    

	readMatures(matFile)
	readSample(sampFile)
	
	#Calculations
	alignShortreads()
	removeUnexpressed()


	return shortread_map, microRNA_map, hairpin_map



def readHairpins(hairpinfile, foldfile):
	"""
	Load hairpin file using biopython.
	Load dot-bracket file.
	"""
	global hairpin_map
	f = open(hairpinfile)
	folds = readFoldstructures(foldfile)
	for record in SeqIO.parse(f, 'fasta'):
		try:
			hairpin_map[record.id] = Hairpin(record.id, record.seq, folds[record.id])
			addRefAsKey(record.id, record.seq)
		except KeyError:
			print('Missing fold structure  for: ' + record.id)
			print record.id in folds
			continue
	# print 'Number of annotated hairpins is %d' % len(hairpin_map)

def readFoldstructures(foldfile):
	"""
	Reads file with fold structures for hairpins, and stores the data in a dict.
	"""
	folds = {}
	f = open(foldfile)
	ref, seq = '', ''
	inProcess = False
	
	for line in f:
		line = line.strip('\r\n') #If file is made in Windows OS
		line = line.strip('\n') 
		if line[0]=='>' and inProcess: 	# each subsequent iteration with > flushes 
										# ref,sec to dictionary
			folds[ref] = seq
			ref = line[1:]
			seq = ''
		elif line[0]=='>': # at first iteration don't add dictionary entry
			ref = line[1:]
			folds[ref] = ''
			inProcess = True
		else:
			seq += line
	folds[ref] = seq
	f.close()
	return folds

def readMatures(filename):
	"""
	Reads input file of mature sequences for reference ID's.
	Format of input lines:
	mmu-let-7g	6	TGAGGTAGTAGTTTGTACAGTT	
	"""
	f = open(filename, 'r')
	seq, ref = '',''
	global microRNA_map
	
	for line in f:
		skip = False
		data = line.split()
		ref = data[0]
		offset = int(data[1])
		sequence = data[2]
		if isInvalidSample(ref, offset, sequence): 
			print 'One mature was skipped!'
			continue
		if ref in hairpin_map:
			#There may be multiple mature seq's pr reference ID, so add a postfix number to mature ID's showing the total count.
			count = len(microRNA_map[ref])
			if count>0:
				for miRNA in microRNA_map[ref]:
					if miRNA.sequence==sequence and miRNA.offset==offset:
						skip = True		
			if not skip:
				newMat = ref + str(count+1)
				newMiRNA = MicroRNA(ref, newMat, 0.0, offset, sequence, findPrime(ref, offset))
				microRNA_map[ref].append(newMiRNA)
	f.close()

	
def addRefAsKey(ref, seq):
	"""
	Adds a new reference ID and it's sequence to all relevant dictionaries.
	"""
	global  miRNA_unique, shortreads
	shortreads[ref] = list()
	shortread_map[ref] = list()
	microRNA_map[ref] = list()

	
def isSeq(seq):
	"""
	Help method for checking if new sequence is valid.
	"""
	allowed = 'ACGT'
	if seq == '' : return False
	for s in seq.upper():
		if s not in allowed:
			return False
	return True

def isInvalidSample(ref, offset, seq):
	"""
	Help method for checking if new sample and reference ID is valid.
	"""
	if ref not in hairpin_map:
		#print 'not in HP map'
		return True
	elif not isSeq(seq):
		#print 'not seq'
		return True
	elif len(hairpin_map[ref].sequence)-offset < len(seq):
		#print 'too long'
		return True
	else:
		return False
	
def countDict (string):
	"""
	Creates dictionaries for indexes and counts of a given sequence's length.
	"""
	l = len(string)
	j=0
	if l%bucketSize > 0: j=1
	counts = dict()
	for i in range (l//bucketSize + j - 1):
		counts[i]=[0]*l	
	return counts
	
def readSample(filename):
	"""
	Reads sample data from given file.
	Updates counts for relevant index ranges in existing reference sequences.
	Format of input lines:
	4-86385	mmu-mir-10b	4	TACCCTGTAGAACCGAATTTGT	0
	"""
	global shortreads, unannotated
	totCount = 0.0
	f = open(filename, 'r')
	ref, seq, offset, count = "", "", 0, 0.0
	
	for line in f:
		sample = line.split()
		count = int(sample[0].split('-')[1])
		totCount += count
	f.close()
	
	
	f = open(filename, 'r')
	for line in f:
		sample = line.split()
		count = int(sample[0].split('-')[1])
		nCount = count/(float(sample[4])+1)
		rpm = (nCount / totCount) * 1000000.0
		
		ref = sample[1]
		offset = int(sample[2])
		seq = sample[3]
		if isInvalidSample(ref, offset, seq): 
			ref, seq, offset, count = '', '', 0, 0.0
			continue
		
		bucketIndex = (len(seq)-1) // bucketSize
			
		#If bucketindex is 2 we have a short read
		if bucketIndex==2: 
			shortreads[ref].append([offset, seq, rpm, ''])
		
		#Save count if sequence is a mature seq
		if bucketIndex in range(3,5):
			annotated = False
			for miRNA in microRNA_map[ref]:
				if miRNA.offset==offset and miRNA.sequence==seq:
					miRNA.count += rpm
					annotated = True
			if not annotated and len(seq)>18:
				if ref not in unannotated:
					unannotated[ref] = [[seq,offset,rpm]]
				else:
					unannotated[ref].append([seq,offset,rpm])


	#Check if any read was higher expressed than the mature sequences for this reference hairpin
	for ref in hairpin_map:
		findUnannotatedSequences(ref)
	f.close()




def findUnannotatedSequences(ref):
	"""
	Alternative to 'checkForNewMatureSeq()' that finds the highest expressed 
	unannotated candidate and compares it with the annotated sequence.
	"""
	global unannotated, microRNA_map, hairpin_map
	
	# Find the highest expressed annotated miRNAs if any
	c3, c5 = 0,0
	annotated3, annotated5 = None, None

	for miRNA in microRNA_map[ref]:
		if miRNA.prime==3:
			c3 = miRNA.count
			annotated3 = miRNA
		elif miRNA.prime==5:
			c5 = miRNA.count
			annotated5 = miRNA


	# If no new unannotated sequences found for the hairpin, save mature sequence if any and set expression level of hairpin if true

	if ref not in unannotated:
		mature3 = annotated3.miRNA_id if c3!=0 else ''
		mature5 = annotated5.miRNA_id if c5!=0 else ''
		mostExpr=0
		if c3>c5:
			mostExpr=3
		elif c5!=0:
			mostExpr=5
		
		for miRNA in microRNA_map[ref]:
			if mature3!='': miRNA.setMature(mature3)
			if mature5!='': miRNA.setMature(mature5)
			if mostExpr!=0: miRNA.setHighestPrime(mostExpr)
		if mostExpr!=0:
			hairpin_map[ref].setExpressed()
		return 0

	highest3, highest5 = ['',0,0], ['',0,0] #[sequence,offset,count]
	for read in unannotated[ref]:
		seq = read[0]
		offset = read[1]
		count = read[2]
		prime = findPrime(ref, offset)
		unannotatedStrand = False
		if (prime==3 and annotated3==None) or (prime==5 and annotated5==None):
			unannotatedStrand = True

		# Check for new 3p sequence
		# Priority: count, offset diff, proximity to 22nt, end offset diff, shortest
		if prime==3 and count>c3:
			if count>highest3[2]:
				highest3 = [seq,offset,count]
			elif count==highest3[2]:
				if unannotatedStrand:
					lendiff_old = abs(22-len(highest3[0]))
					lendiff_new = abs(22-len(seq))
					if lendiff_new<lendiff_old:
						highest3=[seq,offset,count]
					continue
				else:
					startdiff_old = abs(annotated3.offset-highest3[1])
					startdiff_new = abs(annotated3.offset-offset)	
					if startdiff_new<startdiff_old:
						highest3=[seq,offset,count]
					elif startdiff_new==startdiff_old:
						lendiff_old = abs(22-len(highest3[0]))
						lendiff_new = abs(22-len(seq))
						if lendiff_new<lendiff_old:
							highest3=[seq,offset,count]
						elif lendiff_old==lendiff_new:
							enddiff_old = abs((annotated3.offset+len(annotated3.sequence))-(highest3[1]+len(highest3[0])))
							enddiff_new = abs((annotated3.offset+len(annotated3.sequence))-(offset+len(seq)))
							if enddiff_new<enddiff_old:
								highest3 = [seq,offset,count]
							elif enddiff_old==enddiff_new:
								len_old = len(highest3[0])
								len_new = len(seq)
								if len_new<len_old:
									highest3 = [seq,offset,count]
								
		elif prime==5 and count>c5: 
			if count>highest5[2]:
				highest5 = [seq,offset,count]
			elif count==highest5[2]:
				if unannotatedStrand:
					lendiff_old = abs(22-len(highest5[0]))
					lendiff_new = abs(22-len(seq))
					if lendiff_new<lendiff_old:
						highest5=[seq,offset,count]
					continue
				else:
					startdiff_old = abs(annotated5.offset-highest5[1])
					startdiff_new = abs(annotated5.offset-offset)
					if startdiff_new<startdiff_old:
						highest5=[seq,offset,count]
					elif startdiff_new==startdiff_old:
						lendiff_old = abs(22-len(highest5[0]))
						lendiff_new = abs(22-len(seq))
						if lendiff_new<lendiff_old:
							highest5=[seq,offset,count]
						elif lendiff_old==lendiff_new:
							enddiff_old = abs((annotated5.offset+len(annotated5.sequence))-(highest5[1]+len(highest5[0])))
							enddiff_new = abs((annotated5.offset+len(annotated5.sequence))-(offset+len(seq)))
							if enddiff_new<enddiff_old:
								highest5 = [seq,offset,count]
							elif enddiff_old==enddiff_new:
								len_old = len(highest5[0])
								len_new = len(seq)
								if len_new<len_old:
									highest5 = [seq,offset,count]
								
	

	news = 0
	mature3, mature5 = '', ''
	most3, most5 = 0,0
	if highest3[2]>c3:
		newSeq = highest3[0]
		newOffset = highest3[1]
		newCount = highest3[2]
		newPrime = 3
		newID = ref+'3N'
		microRNA_map[ref].append(MicroRNA(ref, newID, newCount, newOffset, newSeq, newPrime))
		mature3 = newID
		most3=newCount 
		news +=1
		#fout.write('%s\t%s\t%d\t%s\n' % (ref, newID, newOffset, newSeq))
	elif highest3[2]<c3:
		mature3 = annotated3.miRNA_id
		most3 = c3
		
	if highest5[2]>c5:
		newSeq = highest5[0]
		newOffset = highest5[1]
		newCount = highest5[2]
		newPrime = 5
		newID = ref+'5N'
		microRNA_map[ref].append(MicroRNA(ref, newID, newCount, newOffset, newSeq, newPrime))
		mature5 = newID
		most5 = newCount
		news += 1
		#fout.write('%s\t%s\t%d\t%s\n' % (ref, newID, newOffset, newSeq))
	elif highest5[2]<c5:
		mature5 = annotated5.miRNA_id
		most5 = c5

	mostExpr = 0
	if most3>most5:
		mostExpr=3
	elif most5>0:
		mostExpr = 5
	else:
		mostExpr = 0

	for miRNA in microRNA_map[ref]:
		if mature3!='': miRNA.setMature(mature3)
		if mature5!='': miRNA.setMature(mature5)
		if mostExpr!=0: miRNA.setHighestPrime(mostExpr)

	if mostExpr!=0:
		hairpin_map[ref].setExpressed()
	
	return



def alignShortreads():
	"""
	To align short reads relative to its reference's most expressed mature sequence for the same strand.
	"""
	global shortreads, shortread_map
	for ref in hairpin_map:

		removed = 0
		for i in range(len(shortreads[ref])):
			read = i - removed
			srOffset = shortreads[ref][read][0]
			srSeq = shortreads[ref][read][1]
			srEnd = srOffset+len(srSeq)
			newPrime = findPrime(ref, srOffset)
			
			mature = None
			for miRNA in microRNA_map[ref]:
				if miRNA.prime==newPrime and miRNA.mature:
					mature = miRNA
			if mature==None:
				continue

			#Might add threshold t:
			#srCount = shortreads[ref][read][2]
			#if srCount<t: continue

			srCount = shortreads[ref][read][2]
			matOffset = mature.offset
			matEnd = matOffset+len(mature.sequence)
			diffStart = srOffset - matOffset
			diffEnd = srEnd-matEnd
			
			# Check if actually a MoR
			if (newPrime==5 and ((srEnd - matOffset)<3)) or (newPrime==3 and ((matEnd - srOffset)<3)):
				continue
							
		
			pos = 'start' if newPrime==3 and abs(diffStart) < abs(diffEnd) or newPrime==5 and abs(diffStart) < abs(diffEnd) else 'end'
			diff = diffEnd if pos=='end' else diffStart
			shortread_map[ref].append(Shortread(ref, srOffset, srSeq, srCount, mature.miRNA_id, pos, newPrime, diff, mature.highest_prime))
			mature.setShortread()
			
	
def isLoop(seq, i):
	chrs = seq[i+1:].replace('.', '')
	if chrs == '': return False
	return True if chrs[0] == ')' else False

def findPrime(ref, offset):
	"""
	Help method for finding what prime the sequence originates from.
	"""
	try:
		seq = hairpin_map[ref].foldstructure
	except IndexError:
		print ('Could not obtain sequence for ref %s, exiting ' % ref)
		return -1
	loops = list()

	for i in range(len(seq)):
		if seq[i] == '(' and isLoop(seq, i):
			i5 = i
			i3 = seq[i:].find(')') + i
			mid = math.ceil((i3+i5)/2.0)
			loops.append(mid)
	
	if len(loops)==0:
		print('Could not find prime as loops is empty for %s' % ref)
		return -1
	
	if len(loops)>1:
		med = len(seq)/2
		mid = min(loops, key=lambda x:abs(x-med))
	else:
		mid = loops[0]

	if offset>mid: return 3
	if offset<=mid: return 5
	
	
def removeUnexpressed():
	"""
	To remove short read statistics for reference hairpins not expressed in sample data.
	"""
	global shortread_map

	unexpressed = 0
	for ref in hairpin_map:
		expr = False

		for miRNA in microRNA_map[ref]:
			if miRNA.mature:
				expr = True
				continue
			if miRNA.count==0:
				miRNA.setUnexpressed()

		if not expr:
			unexpressed += 1
			del shortread_map[ref]
			for miRNA in microRNA_map[ref]:
				miRNA.setUnexpressed()
	
	# print 'Number of unexpressed is: %d' % unexpressed



if __name__ == '__main__':
	"""
	Called when run in command mode.
	Format: 'python findShortreadsMiRNAs Input_hairpins Input_foldsstructures Input_matures Dataset_name Sample1 name1 Sample2 name2 ....'
	"""

	references = sys.argv[1]
	folds = sys.argv[2]
	matures = sys.argv[3]
	total_name = sys.argv[4]

	for i in range(5, len(sys.argv), 2):
		sample = sys.argv[i]
		name = sys.argv[i+1]
		main(references, folds, matures, sample, name, total_name)
	print 'finished :)'
